/*****
 * CKeyEditApp.c
 *
 *	Application methods for the SignWriter key editor.
 *
 *  Copyright  1996 The Center for Sutton Movement Writing.  All rights reserved.
 *
 *****/


#include "CKeyEditApp.h"
#include "Commands.h"
#include "CBartender.h"
#include "CDataFile.h"
#include "CFileWindow.h"
#include "CFileWndKeys.h"
#include "CSymWindow.h"
#include "KeyEditConsts.h"
#include "Keys.h"
#include "Quickdraw.h"
#include <stdlib.h>
#include "Symbols.h"
#include "TBUtilities.h"
#include "TCLUtilities.h"




/*** Global Variables ***/


extern	OSType		 gSignature;
extern	CBureaucrat *gGopher;
extern  CBartender  *gBartender;
extern  CursHandle	 gWatchCursor;
extern  long		 gSleepTime;		/* Max time between events			*/



CSymWindow		*gSymWind;				/* Symbol window access   			*/


CFileWndKeys	*gKeyPanes;				/* Key display access   			*/


Str63			gFileName;				/* Current key file name			*/



void			*gClipRec;				/* Clipboard						*/
void			*gScratchRec;

ObjType			gClipType;
ObjType			gScratchType;



void			*gSignKeyWindow[KEY_CODES];			/* Window management 	*/
void			*gFingerKeyWindow[KEY_CODES];

short			gKeyWindows;



	


/*** Local Constants ***/


#define		kExtraMasters		6
#define		kRainyDayFund		90000
#define		kCriticalBalance	70000
#define		kToolboxBalance		20000




/***
 * IKeyEditApp
 *
 *	Init the application class.  Then read the SignWriter 
 *  symbol file into memory (for later use).
 *
 ***/

void CKeyEditApp::IKeyEditApp(void)

{
	short  x;
	
	
	CApplication::IApplication( kExtraMasters, kRainyDayFund, 
								kCriticalBalance, kToolboxBalance);	
								
/*  The parameters to IApplication are the number of times to call  
    MoreMasters, the total number of bytes of heap space to reserve for                       
    monitoring low memory situations, and the portion of the memory
    reserve to set aside for critical operations and toolbox calls.
    
    Four (4) is a reasonable number of MoreMasters calls,                           
    but you should determine a good number for your application                     
    by observing the heap using Lightsbug,                                              
    TMON, or Macsbug. Set this parameter to zero, give your                         
    program a rigorous work-out, then look at the heap and count                        
    how many master pointer blocks have been allocated. Master                      
    pointer blocks are nonrelocatable and have a size of $100                           
    (hex). You should call MoreMasters at least this many                               
    times -- add a few extra just to be safe. The purpose of all                        
    this preflighting is to prevent heap fragmentation. You                             
    don't want the Memory Manager to call MoreMasters and                           
    create a nonrelocatable block in the middle of your heap. By                        
    calling MoreMasters at the very beginning of the program,                           
    you ensure that these blocks are allocated in a group at the                        
    bottom of the heap. 
                                                                        
    The memory reserve is a safeguard for handling low memory                   
    conditions and is used by the GrowMemory method in                              
    CApplication (check there for more comments). In general,                           
    your program should never request a memory block greater                        
    than this reserve size without explicitly checking in                               
    advance whether there is enough free memory to satisfy the                      
    the request.
                                                                                    
 */ 								
									
	ReadSymFile();
	
	gSymWind 	 = NULL;					// sym window access
	
	gKeyPanes 	 = NULL;					// key display access
	
	
	gClipRec 	 = NULL;					// clipboard management			
	gScratchRec  = NULL;	


	gFileName[0] = 0;						// key file name


	for (x = 0; x < KEY_CODES; x++) {		// window management
	
		gFingerKeyWindow[x] = NULL;		
		gSignKeyWindow[x]   = NULL;	
	}

	gKeyWindows = 0;
}



/***
 * ReadSymFile
 *
 *	Read the SignWriter symbol file.
 *  Exit application if file is missing.
 *
 ***/
 
void CKeyEditApp::ReadSymFile(void)

{
	CDataFile		*theFile = NULL;
	Str63			theName  = SYMFILE_NAME;
	short			symCode;
	Ptr				symBuf = NULL;
	unsigned short 	symBytes;
	short			junk;
	
	
	TRY
	{
			/**
			 ** Create a file object and send it a Specify()
			 **	message to set up the name and volume.
			 **
			 **/
	
		theFile = new CDataFile;
		
		theFile->IDataFile();
		theFile->Specify(CtoPstr(theName), 0);	
	
	
			/**
			 **	Exit app if symbol file missing.
			 **
			 **/
			 
		if (!theFile->ExistsOnDisk()) {
			
			PositionDialog('ALRT', ALRTnoSymFile);	
			junk = Alert(ALRTnoSymFile, NULL);
			
			this->Exit();
			
			exit(0);
			
		}
		
		
			/**
			 **	Show watch cursor during symfile read.
			 **
			 **/
			 
		SetCursor(*gWatchCursor);
		

			/**
			 **	Send the file an Open() message to open it.
			 **
			 **/
	
		theFile->Open(fsRdWrPerm);
		
	
			/**
			 **	Skip number of defined symbols.
			 **
			 **/
			 
		theFile->ReadSome((Ptr) &junk, sizeof(short));
		
		
			/**
			 **	Buffer allocation must succeed.
			 **
			 **/
	
		SetCriticalOperation(TRUE);
		SetAllocation(kAllocCantFail);
		
		
		for (symCode = 1; symCode < SGRID_SYMS; symCode++) {
			
			
			/**
			 **	Read symbol length to determine how many bytes
			 **	to read.  Allocate corresponding buffer and read 
			 ** symbol bytes into it.  Assign buffer to symbol
			 ** data structure.
			 **
			 **/
			 
			theFile->ReadSome((Ptr) &symBytes, sizeof(unsigned short));

			if (symBytes != 0) {
			
				symBuf = NewPtr(symBytes);
				theFile->ReadSome(symBuf, symBytes);
				
				AssignSym(symCode, symBuf);
			}
												
		}	
	
	
			/**
			 **	Close and dispose the file.
			 **
			 **/
	
		theFile->Dispose();
		
	}
	
	CATCH
	{
		 if (theFile) theFile->Dispose();
		 	
		 DisposeSyms();
		 
	}
	ENDTRY; 

}



/***
 * SetUpFileParameters
 *
 *	In this routine, you specify the kinds of files your
 *	application opens.
 *
 *
 ***/

void CKeyEditApp::SetUpFileParameters(void)

{
	inherited::SetUpFileParameters();	/* Be sure to call the default method */

		/**
		 **	sfNumTypes is the number of file types
		 **	your application knows about.
		 **	sfFileTypes[] is an array of file types.
		 **	You can define up to 4 file types in
		 **	sfFileTypes[].
		 **
		 **/

	sfNumTypes = 1;
	sfFileTypes[0] = 'SWKY';
	

		/**
		 **	Although it's not an instance variable,
		 **	this method is a good place to set the
		 **	gSignature global variable. Set this global
		 **	to your application's signature. You'll use it
		 **	to create a file (see CFile::CreateNew()).
		 **
		 **/

	gSignature = 'SWKE';
}



/***
 * DoCommand
 *
 *
 ***/
 
void CKeyEditApp::DoCommand(long theCommand)

{
	short result;
	
	
	switch (theCommand) {
	
		case cmdAbout:
		
					PositionDialog('ALRT', ALRTaboutApp);	
					result = Alert(ALRTaboutApp, NULL);	
					break;
				
				
		default:
					inherited::DoCommand(theCommand);
					break;
	}
}



/***
 *
 * UpdateMenus 
 *
 *   Perform menu management tasks
 *
***/

void CKeyEditApp::UpdateMenus()                    
	 
{
	inherited::UpdateMenus();
	 	
 	gBartender->EnableCmd(cmdAbout);

	gBartender->EnableCmd(cmdNew);
	gBartender->EnableCmd(cmdOpen);
	gBartender->EnableCmd(cmdQuit);
}



/***
 * CreateDocument
 *
 *	The user chose New from the File menu.
 *
 ***/

void CKeyEditApp::CreateDocument()

{
	CFileWindow	*theDocument = NULL;
	
	TRY
	{
		theDocument = new CFileWindow;		
		theDocument->IFileWindow(this, FALSE);
		theDocument->NewFile();
	}
	
	CATCH
	{
		 if (theDocument) theDocument->Dispose();
	}
	ENDTRY;
}



/***
 * OpenDocument
 *
 *	The user chose Open from the File menu.
 *
 *	The macSFReply is a good SFReply record that contains
 *	the name and vRefNum of the file the user chose to
 *	open.
 *
 ***/
 
void CKeyEditApp::OpenDocument(SFReply *macSFReply)

{
	CFileWindow	*theDocument = NULL;
	
	TRY
	{
		theDocument = new CFileWindow;
		theDocument->IFileWindow(this, FALSE);
		theDocument->OpenFile(macSFReply);
	}
	
	CATCH
	{
		 if (theDocument) theDocument->Dispose();
	}
	ENDTRY;
}