/******************************************************************************
 Symbols.c

		Symbol manager for the SignWriter key editor
		
		Copyright  1996 Center for Sutton Movement Writing. All rights reserved.

 ******************************************************************************/
 
 
 
#include "Symbols.h"
#include "Exceptions.h"
#include "Global.h"
#include "LongCoordinates.h"
#include "Memory.h"
#include "KeyEditConsts.h"





/**************************  Symbol data structures  ************************/



typedef struct {		// image 

	Byte			framePtV;
	Byte			framePtH;
	Byte			centerPtV;
	Byte			centerPtH;
	short			numBytes;
	Byte			bitMap[1];
	
} ImageRec, *Image;



typedef struct {		// fill variation  

	Byte			imgClass;
	Byte			numRots;
	Byte			numFlops;
	Byte			numImgs;
	unsigned short	imgInx[1];
	
} FillVarRec, *FillVar;



typedef struct {		// symbol variation 

	Byte			numFills;
	Byte			unused;
	unsigned short	fillInx[1];
	
} SymVarRec, *SymVar;



typedef struct {		// symbol

	Byte			symType;
	Byte			numVars;
	unsigned short	varInx[1];
	
} SymRec, *Sym;



static Ptr 		symTable[SGRID_SYMS] = {NULL};		// symbol table



static Byte		saveBits[IMG_MAP];					// UnDraw bitmap
static BitMap	saveMap;
static Rect		saveSrc;
static Rect     saveDest;





/**************************  Private prototypes  ************************/



void    Translate (ImageClass imgClass, short rot, short flop,
			       short *image, short *transform);




/************************  Symbol initialization  ************************/

	   
void  AssignSym (short symCode, Ptr symPtr)

{
	symTable[symCode] = symPtr;
}


	   
	   
/***************************  Symbol disposal  **************************/

	   
void  DisposeSyms (void)

{
	short symCode;
	
	for (symCode = 1; symCode < SGRID_SYMS; symCode++)
		if (symTable[symCode] != NULL)
			DisposPtr(symTable[symCode]);
}



	   
/****************************  Symbol status  ****************************/


/***
 * SymEmpty
 *
 ***/
 
 Boolean SymEmpty (short symCode)

{
	return symTable[symCode] == NULL;
}




/*************************  Symbol attributes  ***************************/


/***
 * GetSymType
 *
 ***/
 
void GetSymType (short symCode, SymbolType *symType)

{
	*symType = ((Sym) symTable[symCode])->symType;
}



/***
 * GetSymVars
 *
 *
 ***/
 
void GetSymVars (short symCode, short *maxVar)

{
	*maxVar = ((Sym) symTable[symCode])->numVars - 1;	// Sym vars 1-based
}



/***
 * GetFillVars
 *
 ***/
 
void GetFillVars (short symCode, short symVar, short *maxFill)

{
	Ptr	p = symTable[symCode];
	
	*maxFill = ((SymVar) (p + ((Sym) p)->varInx[symVar]))->numFills - 1;    // 1-based
}
	


/***
 * GetAttrs
 *
 ***/	
 			   
void  GetAttrs (short symCode, short symVar, short fillVar,
				ImageClass *imgClass, short *maxRot, short *maxFlop)

{
	Ptr	p = symTable[symCode];

	Ptr p1 = p + ((Sym) p)->varInx[symVar];
	Ptr p2 = p + ((SymVar) p1)->fillInx[fillVar];

	*imgClass = ((FillVar) p2)->imgClass;
	*maxRot   = ((FillVar) p2)->numRots;		// 0-based
	*maxFlop  = ((FillVar) p2)->numFlops;		// 0-based
	
}	




/***********************  Symbol drawing data structures  **********************/


//  Map rotation and flop values into image numbers and image transforms.


typedef enum {n, X, Y, XY} Transform;

			  //     n  no image transform
			  //     X  invert image horizontally
			  //     Y  invert image vertically
			  //    XY  invert image horizontally and vertically
			  
			  
typedef struct {

	Byte	image;
	Byte	transform;	
	
} ImageMap;

	
static ImageMap  mapClass1		[2] = { {0,n},{0,Y}  };
static ImageMap  mapClass1Flop  [2] = { {0,X},{0,XY} };

static ImageMap  mapClass2      [3] = { {0,n},{1,n},{0,Y}  };
static ImageMap  mapClass2Flop  [3] = { {0,X},{1,X},{0,XY} };

static ImageMap  mapClass3A     [8] = { {0,n},{1,n},{2,n},{1,Y},{0,Y},{1,XY},{2,X},{1,X} };
static ImageMap  mapClass3AFlop [8] = { {0,X},{1,X},{2,X},{0,n},{0,n},{0,n}, {0,n},{0,n} };

static ImageMap  mapClass3B     [8] = { {0,n},{1,n}, {2,n},{1,X},{0,X},{1,XY},{2,Y},{1,Y} };
static ImageMap  mapClass3BFlop [8] = { {0,X},{1,XY},{2,Y},{1,Y},{0,n},{0,n}, {0,n},{0,n} };

static ImageMap  mapClass4     [8] = { {0,n},{1,n},{2,n},{3,n},{0,XY},{1,XY},{2,XY},{3,XY} };
static ImageMap  mapClass4Flop [8] = { {0,X},{3,Y},{2,Y},{1,Y},{0,Y}, {3,X}, {2,X}, {1,X}  };

static ImageMap  mapClass5     [8] = { {0,n},{1,n},{2,n},{3,n},{4,n},{3,X},{2,X},{1,X} };
static ImageMap  mapClass5Flop [8] = { {0,X},{1,X},{2,X},{3,X},{4,X},{0,n},{0,n},{0,n} };
	
	
	

//  Map byte value into its bit-wise mirror image.
    
static Byte  flipByte[256] = { 
								0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0, 
								0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0, 
								
								0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8, 
								0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8, 
								
								0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4, 
								0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4, 
								
								0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC, 
								0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC, 
								
								0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2, 
								0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2, 
								
								0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA, 
								0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA, 
								
								0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6, 
								0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6, 
								
								0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE, 
								0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE, 
								
								0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1, 
								0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1, 
								
								0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9, 
								0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9, 
								
								0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5, 
								0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5, 
								
								0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED, 
								0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD, 
								
								0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3, 
								0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3, 
								
								0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB, 
								0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB, 
								
								0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7, 
								0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7, 
								
								0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF, 
								0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
							 };





/****************************  Symbol translation  *****************************/



/***
 * Translate
 *
 ***/
 
void  Translate (ImageClass imgClass, short rot, short flop,
	  		  	 short *image, short *transform)

{
	switch (imgClass) {
	
		case class1:
			
			if (flop) {
				*image     = mapClass1Flop[rot].image;
				*transform = mapClass1Flop[rot].transform;
				
			} else {
				*image     = mapClass1[rot].image;
				*transform = mapClass1[rot].transform;
			}
			break;
		
			
		case class2:
		
			if (flop) {
				*image     = mapClass2Flop[rot].image;
				*transform = mapClass2Flop[rot].transform;
				
			} else {
				*image     = mapClass2[rot].image;
				*transform = mapClass2[rot].transform;
			}
			break;
			
			
		case class3A:
		
			if (flop) {
				*image     = mapClass3AFlop[rot].image;
				*transform = mapClass3AFlop[rot].transform;
				
			} else {
				*image     = mapClass3A[rot].image;
				*transform = mapClass3A[rot].transform;
			}
			break;
			
		
		case class3B:
		
			if (flop) {
				*image     = mapClass3BFlop[rot].image;
				*transform = mapClass3BFlop[rot].transform;
				
			} else {
				*image     = mapClass3B[rot].image;
				*transform = mapClass3B[rot].transform;
			}
			break;
			
			
		case class4:
		
			if (flop) {
				*image     = mapClass4Flop[rot].image;
				*transform = mapClass4Flop[rot].transform;
				
			} else {
				*image     = mapClass4[rot].image;
				*transform = mapClass4[rot].transform;
			}
			break;
			
			
		case class5:
		
			if (flop) {
				*image     = mapClass5Flop[rot].image;
				*transform = mapClass5Flop[rot].transform;
				
			} else {
				*image     = mapClass5[rot].image;
				*transform = mapClass5[rot].transform;
			}
			break;
			
		
		default:
			
			*image     = 0;
			*transform = n;
			break;
						
	}		
}




/***************************   Symbol drawing   ****************************/


/***
 * GetCoords
 *
 ***/
 
void  GetCoords (short symCode, short symVar, short fillVar, short rot, short flop, 
				 LongPt *frame, LongPt *center, Boolean thumbNail)
				   
{
	Ptr			p = symTable[symCode];

	Ptr 		p1 = p + ((Sym) p)->varInx[symVar];
	Ptr 		p2 = p + ((SymVar) p1)->fillInx[fillVar];
	
	ImageClass  imgClass = ((FillVar) p2)->imgClass;
	
	Image		im;
	short		image, transform;
	Point		fram, cent;
	
															// get sym image
	Translate(imgClass, rot, flop, &image, &transform);
	im = (Image) (p + ((FillVar) p2)->imgInx[image]);
	
															// get image coords
	fram.h = im->framePtH;
	fram.v = im->framePtV;
	cent.h = im->centerPtH;
	cent.v = im->centerPtV;

															// adjust coords
	if (transform == X || transform == XY) 
		cent.h = fram.h - cent.h;	
		
	if (transform == Y || transform == XY) 
		cent.v = fram.v - cent.v;
															// scale coords
	if (thumbNail) {
		fram.h /= THUMB_SCALE;   cent.h /= THUMB_SCALE;
		fram.v /= THUMB_SCALE;   cent.v /= THUMB_SCALE;
	}
	
	QDToLongPt(fram, frame);
	QDToLongPt(cent, center);	
	
}



/***
 * DrawSymbol
 *
 ***/
 
void  DrawSymbol (short symCode, short symVar, short fillVar, short rot, short flop,
			      BitMap *img, LongPt *imgPos, 
			      Boolean thumbNail, Boolean canUndraw)
			      
{
	Ptr				p = symTable[symCode];

	Ptr 			p1 = p + ((Sym) p)->varInx[symVar];
	Ptr 			p2 = p + ((SymVar) p1)->fillInx[fillVar];
	
	ImageClass  	imgClass = ((FillVar) p2)->imgClass;
	
	Byte			bitMap[IMG_MAP];
	register short	i, j, count, xSize;
	short			image, transform;
	Rect			source, dest;
	BitMap			map;
	Image			im;

															// get image
	Translate(imgClass, rot, flop, &image, &transform);
	im = (Image) (p + ((FillVar) p2)->imgInx[image]);

															// init draw bitmap
	map.baseAddr = (void *) &bitMap[0];
	map.rowBytes = 2 * ((im->framePtH + 15) / 16);
	SetRect(&map.bounds, 0, 0, im->framePtH, im->framePtV);
	
	count = im->numBytes;
	xSize = map.rowBytes;
	
															// transform image
	switch (transform) {
	
	
		case n:												//  no transform
		
			for (i = 0; i < count; i++)
				bitMap[i] = im->bitMap[i];
						
			break;
		
		
		case X:				  								//  invert horizontally
	
			for (j = 0; j < count; j += xSize) 
				for (i = 0; i < xSize; i++)
					bitMap[i+j] = flipByte[im->bitMap[xSize-1-i+j]];	
					
			break;
	
	
		case Y:			  									//  invert vertically

			for (j = 0; j < count; j += xSize) 
				for (i = 0; i < xSize; i++)
					bitMap[i+j] = im->bitMap[i+count-xSize-j];	
			
			break;
		
		
		case XY:			  								//  invert horiz. and vert.
	
			for (i = 0; i < count; i++)
				bitMap[i] = flipByte[im->bitMap[count-1-i]];	
		
			break;			
	}
		
															//  init image source
	source = map.bounds;
		
	if (transform == X || transform == XY) 
		OffsetRect(&source, (xSize * 8) - im->framePtH, 0);	
		
															// init image dest
	dest = map.bounds;
	
	if (thumbNail) {	
		dest.bottom /= THUMB_SCALE;
		dest.right  /= THUMB_SCALE;
	}
																	
	saveSrc = dest;

	OffsetRect(&dest, imgPos->h, imgPos->v);	
	saveDest = dest;

															// save existing bits
	if (canUndraw) {		
	
		saveMap = map;
		saveMap.baseAddr = (void *) &saveBits[0];

		CopyBits(img, &saveMap, &dest, &saveSrc, srcCopy, NULL);
		
	}
		
															// draw symbol														
	CopyBits(&map, img, &source, &dest, srcOr, NULL);
			
}				 



/***
 * UndrawSymbol
 *
 ***/
 
void  UndrawSymbol (BitMap *img)

{		
	CopyBits(&saveMap, img, &saveSrc, &saveDest, srcCopy, NULL);
}
